#     
#     File name: Functions.s
#
#     Defines functions
#
#     ================================================
#
#     Compute B-spline regression matrix.
#
#     Arguments
#
#     X. = x-domain vector
#     XL. = left of x-domain, XR. = right of x-domain
#     NDX. = number of intervals in domain
#     BDEG. = degree of B-spline (quadratic = 2, etc)
#	
bspline <- function(X., XL., XR., NDX., BDEG.){
 dx <- (XR. - XL.)/NDX.
 knots <- seq(XL. - BDEG.*dx, XR. + BDEG.*dx, by=dx)
 B <- spline.des(knots,X.,BDEG.+1,0*X.)$design
 B
}
#
#     Computes B and D'D
#
#     Arguments 
#
#     As bspline( ).  Also 
#     PORD. = order of the penalty
#
setup.P <- function(X., XL., XR., BDEG., PORD., NDX.){
  B <- bspline(X., XL., XR., NDX., BDEG.)
  D <- diag(ncol(B))
  for (k in 1:PORD.) D <- diff(D)
  DtD <- t(D) %*% D
  return(B, DtD)
}
#
#     Update function: this performs a single iteration of (3.4)
#
#     Arguments
#
#     X. = Regression matrix
#     V. = data present/missing vector
#     OFF. = offset vector
#     D. = claims vector
#     A. = current estimate of regression coefficients
#     P. = penalty matrix
#
Update <- function(X., V., OFF., D., A., P.){
  Ita <- OFF. + X. %*% A.
  Mu <- c(exp(Ita))
  Wt <- Mu * V.
  Bt.W.B <- t(X.) %*% (Wt * X.)
  Rhs <- Bt.W.B %*% A. + t(X.) %*% (V. * (D. - Mu))
  a <- solve(Bt.W.B + P., Rhs)
  a
 }
#
#      Estimate regression coefficients.  This calls the update function
#      until convergence.  The user can set convergence criterion and
#      maximum number of iterations.  Monitoring of output can be
#      switched off/on.
#
#      Arguments
#
#      As Update().  Also
#      A.INIT = initial estimate of a
#      TOL. = tolerance, set to 10^(-8) or otherwise
#      MONITOR. = monitoring, FALSE = off, TRUE = on
#      MAX.ITER. = maximum no of iterations, set to 20 or otherwise
#
Estimate <- function(A.INIT., X., V., OFF., D., P.,
                     TOL., MONITOR., MAX.ITER.){
  a <- A.INIT.
  a.old <- 10
  iter <- 0
  tol = 1
  if (MONITOR. == TRUE) cat("Iter     a[1]            tol", "\n")
  while(tol > TOL. && iter < MAX.ITER.) {
    iter <- iter + 1
    a <- Update(X., V., OFF., D., a, P.)
    tol <- max(abs(a - a.old))/mean(abs(a))
    a.old <- a
    if (MONITOR. == TRUE) cat(iter, a[1], tol, "\n")
  }
  if (iter == MAX.ITER.) warning("Failed to converge")
  a
}
#
#       Main function: calculates
#
#          Calc$a   = estimates of regression coefficients
#          Calc$Hazard = the hazard function = Ba
#          Calc$Tr  = trace of hat matrix
#          Calc$Dev = deviance
#          Calc$Bic = BIC (Bayesian Information Criterion)
#          Calc$BtWBplusP = B'WB + P (needed for SE's)
#
#       Arguments
#
#       A.INIT., X., V., OFF., D., TOL., MONITOR., MAX.ITER. as above.
#       Also
#       P1. = penalty matrix for age
#       LAMBDA1. = smoothing parameter for age
#       P2. = penalty matrix for year
#       LAMBDA2. = smoothing parameter for year
#       TRACE. = used in conjunction with minimization
#       If ms() is used set to TRUE, otherwise FALSE
#
#       Note: lines 1 and 2 of the function make sure that the
#       lambda values do not go negative.
#
Calc <- function(A.INIT., X., V., OFF., D.,
                 P1., LAMBDA1., P2., LAMBDA2.,
                 TOL., MONITOR., MAX.ITER., TRACE.){
  lambda1 <- abs(LAMBDA1.)
  lambda2 <- abs(LAMBDA2.)
  P <- lambda1 * P1. + lambda2 * P2.
  a <- Estimate(A.INIT., X., V., OFF., D., P,
                TOL., MONITOR., MAX.ITER.)
  Ita <- OFF. + X. %*% a
  Mu <- c(exp(Ita))
  Wt <- Mu * V.
  Bt.W.B <- t(X.) %*% (Wt * X.)
  BtWBplusP <- Bt.W.B + P
  Rhs <- Bt.W.B %*% a + t(X.) %*% (V. * (D. - Mu))
  a <- solve(BtWBplusP, Rhs)
  Tr <- sum(diag(solve(BtWBplusP, Bt.W.B)))
  y.init <- D.
  y.init[D.==0] <- 10^(-4)
  Dev <- 2*sum( V. * D.*log(y.init/Mu) )
  Bic <- Dev + log(sum(V.)) * Tr
  Hazard <- Ita - OFF.
  if (TRACE. == TRUE) cat(lambda1, lambda2, Bic, "\n")
  return(a, Hazard, Tr, Dev, Bic, BtWBplusP)
}
#
#       Grid search for age
#
#       Searches for minimum value of lambda for age over a grid
#       for fixed value of lambda for year.  Grid size is 5 (although
#       this could be changed).
#
#       Arguments
#
#       A.INIT., X., V., E., D., P1., P2., TOL., MONITOR., MAX.ITER.,
#       TRACE. as above. Also
#       LEFT. = left end of grid
#       RIGHT. = right end of grid
#       SCALE. = 0 for log search, 1 for linear scale
#       LAMBDA.YEAR. = fixed value of lambda for year
#
Grid.search.age <- function(LEFT., RIGHT., SCALE., LAMBDA.YEAR.,
                            A.INIT., X., V., OFF., D., P1., P2.,
                            TOL., MONITOR., MAX.ITER., TRACE.){
  lambda.list <- seq(LEFT., RIGHT., length = 5)
  if (SCALE. == 0) lambda.list <- 10^lambda.list
  Tr.list <- NULL
  Dev.list <- NULL
  Bic.list <- NULL
  cat("Iter           Trace          Bic", "\n")
  for (i in 1:length(lambda.list)){
    lambda.age <- lambda.list[i]
    Output <- Calc(A.INIT., X., V., OFF., D.,
                   P1.,lambda.age, P2., LAMBDA.YEAR.,
                   TOL., MONITOR., MAX.ITER., TRACE.)
    Tr.list <- c(Tr.list, Output$Tr)
    Dev.list <- c(Dev.list, Output$Dev)
    Bic.list <- c(Bic.list, Output$Bic)
    cat(i, Output$Tr, Output$Bic, "\n")
  }
  if (SCALE. == 0) lambda.list <- log(lambda.list)/log(10)
  cbind(lambda.list, Tr.list, Dev.list, Bic.list)
}
#
#      Grid search for year: as grid search for age but with
#
#      LAMBDA.AGE. = fixed value of lambda for age
#
Grid.search.year <- function(LEFT., RIGHT., SCALE., LAMBDA.AGE.,
                             A.INIT., X., V., OFF., D., P1., P2.,
                             TOL., MONITOR., MAX.ITER., TRACE.){
  lambda.list <- seq(LEFT., RIGHT., length = 5)
  if (SCALE. == 0) lambda.list <- 10^lambda.list
  Tr.list <- NULL
  Dev.list <- NULL
  Bic.list <- NULL
  cat("Iter      Trace        Bic", "\n")
  for (i in 1:length(lambda.list)){
    lambda.year <- lambda.list[i]
    Output <- Calc(A.INIT., X., V., OFF., D.,
                   P1., LAMBDA.AGE., P2., lambda.year,
                   TOL., MONITOR., MAX.ITER., TRACE.)
    Tr.list <- c(Tr.list, Output$Tr)
    Dev.list <- c(Dev.list, Output$Dev)
    Bic.list <- c(Bic.list, Output$Bic)
    cat(i, Output$Tr, Output$Bic, "\n")
  }
  if (SCALE. == 0) lambda.list <- log(lambda.list)/log(10)
  cbind(lambda.list, Tr.list, Dev.list, Bic.list)
}


